from django.shortcuts import render, redirect
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import FormView, View

from wspay.conf import settings
from wspay.forms import (
    UnprocessedPaymentForm, WSPaySignedForm, WSPayErrorResponseForm, WSPaySuccessResponseForm,
    WSPayCancelResponseForm,
)
from wspay.models import WSPayRequestStatus
from wspay.services import process_input_data, process_response_data, verify_response


class ProcessView(FormView):
    """Receive payment data and prepare it for WSPay."""

    form_class = UnprocessedPaymentForm
    template_name = 'wspay/error.html'

    def form_valid(self, form):
        wspay_form = WSPaySignedForm(
            process_input_data(form.cleaned_data.copy(), self.request)
        )
        return render(
            self.request,
            'wspay/wspay_submit.html',
            {'form': wspay_form, 'submit_url': settings.WS_PAY_PAYMENT_ENDPOINT}
        )


class PaymentStatus:
    SUCCESS = 'success'
    ERROR = 'error'
    CANCEL = 'cancel'


class ProcessResponseView(View):
    """Handle success, error and cancel."""

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        form_class, request_status, redirect_url = self._unpack_response_status(kwargs['status'])

        data = request.POST if request.method == 'POST' else request.GET
        process_response_data(
            verify_response(form_class, data),
            request_status
        )

        return redirect(redirect_url)

    def _unpack_response_status(self, status):
        assert status in [PaymentStatus.SUCCESS, PaymentStatus.ERROR, PaymentStatus.CANCEL]
        if status == PaymentStatus.SUCCESS:
            form_class = WSPaySuccessResponseForm
            request_status = WSPayRequestStatus.COMPLETED
            redirect_url = settings.WS_PAY_SUCCESS_URL
        elif status == PaymentStatus.CANCEL:
            form_class = WSPayCancelResponseForm
            request_status = WSPayRequestStatus.CANCELLED
            redirect_url = settings.WS_PAY_CANCEL_URL
        else:
            form_class = WSPayErrorResponseForm
            request_status = WSPayRequestStatus.FAILED
            redirect_url = settings.WS_PAY_ERROR_URL

        return form_class, request_status, redirect_url


class TestView(FormView):
    """Simple View to test the ProcessView."""

    template_name = 'wspay/test.html'
    form_class = UnprocessedPaymentForm



from appconf import AppConf

from django.conf import settings  # noqa


class WSPayAppConf(AppConf):

    SUCCESS_URL = '/'
    ERROR_URL = '/'
    CANCEL_URL = '/'
    BASE_URL = None,
    DEVELOPMENT = None
    PAYMENT_ENDPOINT = None
    VERSION = '2.0'

    class Meta:
        prefix = 'ws_pay'
        required = ['SHOP_ID', 'SECRET_KEY']

    def configure(self):
        development = self.configured_data['DEVELOPMENT']

        # Set payment endpoint based on WS_PAY_DEVELOPMENT if it is not explicitly
        # set.
        if not self.configured_data['PAYMENT_ENDPOINT']:
            if development:
                payment_endpoint = 'https://formtest.wspay.biz/authorization.aspx'
            else:
                payment_endpoint = 'https://form.wspay.biz/authorization.aspx'
            self.configured_data['PAYMENT_ENDPOINT'] = payment_endpoint

        return self.configured_data

    def configure_development(self, value):
        return settings.DEBUG if value is None else value



import pytest
import responses
import requests

from django.urls import reverse
from django.test.client import RequestFactory

from wspay.forms import UnprocessedPaymentForm, WSPaySignedForm
from wspay.models import WSPayRequest
from wspay.services import process_input_data, generate_signature

from django.conf import settings


def test_incoming_data_form():
    """Test the form that receives cart and user details."""
    form = UnprocessedPaymentForm()
    assert form.is_valid() is False
    form = UnprocessedPaymentForm({'user_id': 1, 'cart_id': 1, 'price': 1})
    assert form.is_valid()


@pytest.mark.django_db
def test_wspay_encode():
    """Test the processing function which prepares the data for WSPay."""
    assert settings.WS_PAY_SHOP_ID == 'ljekarnaplus'
    assert settings.WS_PAY_SECRET_KEY == '123456'

    return_data = {
        'ShopID': settings.WS_PAY_SHOP_ID,
        'Version': settings.WS_PAY_VERSION,
        'TotalAmount': '10,00',
        'ReturnURL': (
            'http://testserver' + reverse('wspay:process-response', kwargs={'status': 'success'})
        ),
        'CancelURL': (
            'http://testserver' + reverse('wspay:process-response', kwargs={'status': 'cancel'})
        ),
        'ReturnErrorURL': (
            'http://testserver' + reverse('wspay:process-response', kwargs={'status': 'error'})
        ),
        'ReturnMethod': 'POST',
    }

    incoming_form = UnprocessedPaymentForm({'cart_id': 1, 'price': 10})
    if (incoming_form.is_valid()):
        form_data = process_input_data(
            incoming_form.cleaned_data.copy(), RequestFactory().get('/')
        )

    req = WSPayRequest.objects.get()
    return_data['ShoppingCartID'] = str(req.request_uuid)
    return_data['Signature'] = generate_signature(
        [settings.WS_PAY_SHOP_ID, str(req.request_uuid), '1000']
    )

    assert return_data == form_data


@pytest.mark.django_db
def test_wspay_form():
    """Test the form that is used to make a WSPay POST request."""
    form = WSPaySignedForm()
    assert form.is_valid() is False

    incoming_form = UnprocessedPaymentForm({'user_id': 1, 'cart_id': 1, 'price': 1})
    if (incoming_form.is_valid()):
        form_data = process_input_data(
            incoming_form.cleaned_data.copy(),
            RequestFactory().get('/')
        )

    form = WSPaySignedForm(form_data)
    assert form.is_valid()
    form = form.cleaned_data

    responses.add(responses.POST, 'https://formtest.wspay.biz/authorization.aspx', status=200)
    response = requests.post('https://formtest.wspay.biz/authorization.aspx', form)
    assert response.status_code == 200



from decimal import setcontext, Decimal, BasicContext
import json
import hashlib

from django.core.exceptions import ValidationError
from django.urls import reverse

from wspay.conf import settings
from wspay.models import WSPayRequest
from wspay.signals import pay_request_created, pay_request_updated

EXP = Decimal('.01')
setcontext(BasicContext)


def process_input_data(input_data, request, user=None):
    """Process incoming data and prepare for POST to WSPay."""
    wspay_request = WSPayRequest.objects.create(
        cart_id=input_data['cart_id'],
    )
    # Send a signal
    pay_request_created.send_robust(WSPayRequest.__class__, instance=wspay_request)

    input_data['cart_id'] = str(wspay_request.request_uuid)

    price = input_data['price']
    assert price > 0, 'Price must be greater than 0'
    total_for_sign, total = build_price(price)

    signature = generate_signature(
        [settings.WS_PAY_SHOP_ID, input_data['cart_id'], total_for_sign]
    )

    return_data = {
        'ShopID': settings.WS_PAY_SHOP_ID,
        'ShoppingCartID': input_data['cart_id'],
        'Version': settings.WS_PAY_VERSION,
        'TotalAmount': total,
        'Signature': signature,
        'ReturnURL': request.build_absolute_uri(
            reverse('wspay:process-response', kwargs={'status': 'success'})
        ),
        'CancelURL': request.build_absolute_uri(
            reverse('wspay:process-response', kwargs={'status': 'cancel'})
        ),
        'ReturnErrorURL': request.build_absolute_uri(
            reverse('wspay:process-response', kwargs={'status': 'error'})
        ),
        'ReturnMethod': 'POST',
    }
    if input_data.get('first_name'):
        return_data['CustomerFirstName'] = input_data['first_name']
    if input_data.get('last_name'):
        return_data['CustomerLastName'] = input_data['last_name']
    if input_data.get('address'):
        return_data['CustomerAddress'] = input_data['address']
    if input_data.get('city'):
        return_data['CustomerCity'] = input_data['city']
    if input_data.get('zip_code'):
        return_data['CustomerZIP'] = input_data['zip_code']
    if input_data.get('country'):
        return_data['CustomerCountry'] = input_data['country']
    if input_data.get('email'):
        return_data['CustomerEmail'] = input_data['email']
    if input_data.get('phone'):
        return_data['CustomerPhone'] = input_data['phone']

    return return_data


def verify_response(form_class, data):
    """Verify validity and authenticity of wspay response."""
    form = form_class(data=data)
    if form.is_valid():
        signature = form.cleaned_data['Signature']
        param_list = [
            settings.WS_PAY_SHOP_ID,
            data['ShoppingCartID'],
            data['Success'],
            data['ApprovalCode']
        ]
        expected_signature = generate_signature(param_list)
        if signature != expected_signature:
            raise ValidationError('Bad signature')

        return form.cleaned_data

    raise ValidationError('Form is not valid')


def process_response_data(response_data, request_status):
    """Update corresponding WSPayRequest object with response data."""
    wspay_request = WSPayRequest.objects.get(
        request_uuid=response_data['ShoppingCartID'],
    )
    wspay_request.status = request_status.name
    wspay_request.response = json.dumps(response_data)
    wspay_request.save()

    # Send a signal
    pay_request_updated.send_robust(
        WSPayRequest.__class__,
        instance=wspay_request,
        status=request_status
    )

    return wspay_request


def generate_signature(param_list):
    """Compute the signature."""
    result = []
    for x in param_list:
        result.append(x)
        result.append(settings.WS_PAY_SECRET_KEY)
    return compute_hash(''.join(result))


def compute_hash(signature):
    """Compute the hash out of the given values."""
    return hashlib.sha512(signature.encode()).hexdigest()


def build_price(price):
    """
    Round to two decimals and return the tuple containing two variations of price.

    First element of the tuple is an int repr of price as as str 123.45 => '12345'
    Second element is a str that is a properly formatted price 00123.451 => '123,45'
    """
    rounded = price.quantize(EXP)
    _, digits, exp = rounded.as_tuple()

    result = []
    digits = list(map(str, digits))
    build, next = result.append, digits.pop

    for i in range(2):
        build(next() if digits else '0')
    build(',')
    if not digits:
        build('0')

    while digits:
        build(next())

    return str(int(rounded * 100)), ''.join(reversed(result))
